/*
Copyright (C) 2012-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
// command_sys.c -- command interpreter and processor

#include "core.h"
#include "corex.h" // Not part of core
#include "command_sys.h" // Courtesy include

///////////////////////////////////////////////////////////////////////////////
//  LOCAL SETUP
///////////////////////////////////////////////////////////////////////////////

#define MLOCAL(_m) struct mlocal *_m = (struct mlocal *) me->_local;

// Begin ...

static const char * const _tag = "cmdsys";  	// TAG
#define mobj_t cmdsys_t						// object type
#define mlocal _cmdsys_local					// private data struct
#define mitem_t _cmdsys_item_t
#define mitem_s _cmdsys_item_s
#define mtracker_s _cmdsys_track_t
#define mtracker_t _cmdsys_track_s

typedef struct mitem_s
{
	struct mitem_s *next, *prev;
	
	sysobj_t	*object; // Object that owns the command as we need an instance
	cmdinfo_t	*cmdinfo;
	const char  *name; // allocated
	
} mitem_t;

typedef struct mtracker_s
{
	mitem_t		*head;
	mitem_t		*tail;
	int			count;
} mtracker_t;

struct mlocal							// private data
{
	mtracker_t	tracker;	// Tracks subsystems
	memchain_t	*memchain;	// Stores our memory allocations for easy disposal
	sizebuf_t	cmd_buffer;
};


///////////////////////////////////////////////////////////////////////////////
//  INTERNAL FUNCTIONS
///////////////////////////////////////////////////////////////////////////////

static void sLinkAlpha (mtracker_t *track, mitem_t *newitem)
{
	// Four scenarios:
	// 1) First item.  Take head and tail.
	// 2) Beat head.  Take head, bump one.
	// 3) Beat nothing.  Become tail, trail one.
	// 4) Insertion.  Previous, <--prev(me)next-->, Next are modified (4 changes)


	// First see if we are the top item
	if (track->head == NULL)
	{
		// No items, assume head and tail, no bump and no insertion
		track->head = newitem;
		track->tail = newitem;
	}
	else if (strcasecmp (track->head->name, newitem->name) > 0 )
	{
		// Beat head and are not the tail 
		newitem->next = track->head; // My next
		track->head->prev = newitem; // Bumped item previous
		track->head = newitem;		 // No previous, just take head
	}
	else
	{
		// Not top ... run through list and stop at end or when we alphabetically beat an item
		mitem_t *cursor;

		for (cursor = track->head; cursor; cursor = cursor->next )
			if (strcasecmp(cursor->name, newitem->name)  > 0)
				break;

		// This returns the worst item we do NOT alphabetically beat, so it cannot return NULL
		// So this returns the item we are *AFTER*
		if (cursor == NULL)
		{
			// We have tail
			track->tail->next = newitem; // Item before's next is me
			newitem->prev = track->tail; // My previous is old tail
			track->tail = newitem;		 // I am tail.
		}
		else
		{
			// We beat something
			newitem->next = cursor;			// My next is this item we beat
			newitem->prev = cursor->prev;   // My previous is the beaten item's previous
			cursor->prev->next = newitem;	// The item before the beaten item's next is now me
			cursor->prev = newitem;			// Beaten item's previous is me.
		}
	}

	track->count ++;
}


static void sUnlink (mtracker_t *track, mitem_t *item)
{
	if (item->prev) 
		item->prev->next = item->next;
	else track->head = NULL; // I was head

	if (item->next) 
		item->next->prev = item->prev;
	else track->tail = NULL; // I was tail

	track->count --;
}


static mitem_t *sFindLink (mtracker_t *track, cmdinfo_t *cmdinfo)
{
	mitem_t *cur;

	for (cur = track->head; cur; cur = cur->next)
		if (cur->cmdinfo == cmdinfo)
			return cur;

	return NULL;
}

static mitem_t *sFindLinkName (mtracker_t *track, const char *s)
{
	mitem_t *cur;

	for (cur = track->head; cur; cur = cur->next)
		if (!strcasecmp(cur->name, s))
			return cur;

	return NULL;
}

///////////////////////////////////////////////////////////////////////////////
//  PUBLIC INFORMATIONAL OBJECTS
///////////////////////////////////////////////////////////////////////////////

static void Report (mobj_t *me)
{
	MLOCAL(m)

	mitem_t *cur;
	int count;

	Core_Printf ("Command Register\n-----\n");
	// Print A to Z
	for (cur = m->tracker.head, count = 0; cur; cur = cur->next, count ++)
		Core_Printf ("%03i: (object: %p) Type: \"%s\"\n", count, cur->cmdinfo, cur->name);

	Core_Printf ("--Found %i -----\n", count);
}

static int Report_Cmd (sysobj_t *me, lparse_t *line)
{
	Report ((mobj_t *) me);
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//  PUBLIC OBJECT FUNCTIONS: COMMAND ADD / REMOVE
///////////////////////////////////////////////////////////////////////////////

static void *sCommandRemoveItem (mobj_t *me, mitem_t *cur)
{
	MLOCAL (m)

	// Verify
	if (strcmp (cur->name, cur->cmdinfo->name))
		Core_Error ("Remove: Command name does not match");

	// Unlink from our list
	sUnlink (&m->tracker, cur);

	// Deallocate our copy of the name and cmdinfo
	m->memchain->Free (m->memchain, (void *)cur->name);
	m->memchain->Free (m->memchain, (void *)cur->cmdinfo);

	// Deallocate link item it using memory chain
	m->memchain->Free (m->memchain, cur);

	return NULL;
}

// Close one subsystem
static void *CommandRemove (mobj_t *me, cmdinfo_t *cmdinfo)
{	
	// If object is NULL that would be disturbing
	if (cmdinfo)
	{
		MLOCAL(m)
		
		// Locate it in the list
		mitem_t *cur = sFindLink (&m->tracker, cmdinfo);

		if (!cur)
			Core_Error ("Unable to remove command because not registered!");
		
		return sCommandRemoveItem (me, cur);
	}
	return NULL;
}

// Find the item containing a command
static mitem_t *sCommandFind (mobj_t *me, const char *name)
{	
	MLOCAL(m)
	
	// Locate it in the list
	mitem_t *cur = sFindLinkName (&m->tracker, name);
	if (cur)
		return cur;

	return NULL;
}


static void *CommandAdd (mobj_t *me, sysobj_t *object, cmdinfo_t *cmdinfo)
{
	// Only if registration succeeded (failed is NULL)
	MLOCAL(m)

	// Use the memory chain to allocate a subsystems item item
	mitem_t *item = m->memchain->Calloc (m->memchain, sizeof(mitem_t), 1);

	// Fill in the object
	item->name = m->memchain->Strdup (m->memchain, cmdinfo->name);
	item->cmdinfo = m->memchain->Memdup (m->memchain, cmdinfo, sizeof(cmdinfo_t));
	item->object = object; // Owner

	// Link it in
	sLinkAlpha (&m->tracker, item);

	return object;
}

static void *CommandAddX (mobj_t *me, const char *name, int (*command_fn) (sysobj_t *, lparse_t *), const char *helptext)
{
	cmdinfo_t cmdinfo = {name, command_fn, helptext};

	return CommandAdd (me, NULL, &cmdinfo);
}

///////////////////////////////////////////////////////////////////////////////
//  PUBLIC OBJECT FUNCTIONS: COMMAND BUFFER
///////////////////////////////////////////////////////////////////////////////

static void InsertText (mobj_t *me, const char *text);

static int ExecuteString (mobj_t *me, const char *text)
{
	MLOCAL(m)
	
	lparse_t line = Line_Parse (text);

	if (line.count)
	{
		mitem_t *item = sFindLinkName (&m->tracker, line.args[0]);
		
		// 	If we found command, run it and return it.
		if (item)
			return item->cmdinfo->command_fn (item->object, &line);

		Core_Printf ("Unknown command \"%s\"\n", line.args[0]);
	}
	
	// No arguments or command not found
	return (1);  // Return failure (1)
}


// Execute the command buffer
static void Execute (mobj_t *me)
{
	MLOCAL(m)

	char text[MAX_CMD_256];

	while (Buffer_GetLine (&m->cmd_buffer, text, sizeof(text)) )
	{
// execute the command line
		ExecuteString (me, text);
	}
}

// Adds command text at the end of the buffer
static void AddText (mobj_t *me, const char *text)
{
	MLOCAL(m)
	
	int	len = strlen (text);

	if (m->cmd_buffer.cursize + len >= m->cmd_buffer.maxsize)
	{
		Core_Printf ("AddText: overflow\n");
		return;
	}

	Buffer_Write (&m->cmd_buffer, text, len);
}

// Adds command text immediately after the current command
// Adds a \n to the text
static void InsertText (mobj_t *me, const char *text)
{
	MLOCAL(m)

	char	*temp = NULL;
	int		templen = m->cmd_buffer.cursize;

// copy off any commands still remaining in the exec buffer
	if (templen)
	{
		temp = (char *) core_calloc (templen, 1);
		memcpy (temp, m->cmd_buffer.data, templen);
		Buffer_Clear (&m->cmd_buffer);
	}

// add the entire text of the file
	AddText (me, text);
	Buffer_Write (&m->cmd_buffer, "\n", 1);

// add the copied off data
	if (templen)
	{
		Buffer_Write (&m->cmd_buffer, temp, templen);
		core_free (temp);
	}
}


static cmdinfo_t commands[] =
{
	"commands", Report_Cmd, "List of commands available"
}; ARRAY_SIZE_STATIC(commands);


static cbool Initialize (mobj_t *me)
{
	MLOCAL(m)
	
	int i;

	// Anything we do here
	m->memchain = Memchain_Instance ();

	// Add ourselves a command
	for (i = 0; i < _numcommands; i ++)
		CommandAdd (me, (sysobj_t *)me, &commands[i]);

	// Allocate command buffer
	Buffer_Alloc (&m->cmd_buffer, 8192); // Why not?

	return true;
}


///////////////////////////////////////////////////////////////////////////////
//  PUBLIC GLOBAL FUNCTIONS:  Creation and freeing of object
///////////////////////////////////////////////////////////////////////////////


// Shutdown and Free
static void *Shutdown (mobj_t *me)
{
	MLOCAL(m)

	mitem_t *cur;

	// Shutdown command buffer
	Buffer_Free (&m->cmd_buffer);

	// Shutdown, works A to Z
	for (cur = m->tracker.head; cur; cur = cur->next)
		sCommandRemoveItem (me, cur);

	m->memchain->Shutdown (m->memchain);

	core_free (me->_local);
	core_free (me);

	return NULL;
}


cmdsys_t *CmdSys_Instance (void)
{
	cbool result;

	mobj_t *nobj = core_calloc (sizeof(mobj_t), 1);
	nobj->_local = core_calloc (sizeof(struct mlocal), 1);

	// Hook up functions
	#define FUNCTION_HOOKUP(_x) nobj->_x = _x

	FUNCTION_HOOKUP(CommandAdd);
	FUNCTION_HOOKUP(CommandAddX);
	FUNCTION_HOOKUP(CommandRemove);
	FUNCTION_HOOKUP(Report);

	FUNCTION_HOOKUP(AddText);
	FUNCTION_HOOKUP(InsertText);
	FUNCTION_HOOKUP(ExecuteString);
	FUNCTION_HOOKUP(Execute);

	#undef FUNCTION_HOOKUP

	// Run initializer
	result = Initialize (nobj);

	if (!result)
	{
		nobj = Shutdown (nobj);
	}

	// Required information
	OBJ_REQUIRED_HOOKUP(nobj) // Sets parent, _tag, Shutdown

	return nobj;
}

#undef mobj_t 
#undef mlocal

	