/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2010-2014 Quakespasm 
Copyright (C) 2012-2014 Baker

This program is core_free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// image.c -- image manipulation and loading

#define CORE_LOCAL

#include "core.h"
#include "image.h"



///////////////////////////////////////////////////////////////////////////////
//  IMAGE: OPERATIONS
///////////////////////////////////////////////////////////////////////////////

void Image_Flip_Buffer (byte *buffer, int columns, int rows, int bytes_per_pixel)
{
	int		bufsize = columns * bytes_per_pixel; // bufsize = widthBytes;
	byte*	tb1 = core_malloc (bufsize); // Flip buffer
	byte*	tb2 = core_malloc (bufsize); // Flip buffer2
	int		i, offset1, offset2;

	for (i = 0; i < (rows + 1) / 2; i ++)
	{
		offset1 = i * bufsize;
		offset2 = ((rows - 1) - i) * bufsize;

		memcpy(tb1,				buffer + offset1, bufsize);
		memcpy(tb2,				buffer + offset2, bufsize);
		memcpy(buffer+offset1,	tb2,			  bufsize);
		memcpy(buffer+offset2,	tb1,			  bufsize);
	}

	core_free (tb1);
	core_free (tb2);
	return;
}

///////////////////////////////////////////////////////////////////////////////
//  IMAGE: MISCELLANEOUS
///////////////////////////////////////////////////////////////////////////////

int Image_Power_Of_Two_Size (int n)
{
	int upsized_pow2;
	for (upsized_pow2 = 1 ; upsized_pow2 < n ; upsized_pow2 <<= 1) //  <<= 1 is a bitwise double, same as "times two"
		;

	Core_DPrintf ("Upsized %i to %i\n", n, upsized_pow2);
	return upsized_pow2;
}

///////////////////////////////////////////////////////////////////////////////
//  IMAGE: GAMMA 
///////////////////////////////////////////////////////////////////////////////

void Image_Build_Gamma_Table (float g, float c, byte gammatable_256[])
{
	int	i, inf;

	g = CLAMP(0.3, g, 3);
	c = CLAMP(1, c, 3);

	if (g == 1 && c == 1)
	{
		for (i = 0 ; i < 256 ; i++)
			gammatable_256[i] = i;
		return;
	}

	for (i = 0 ; i < 256 ; i++)
	{
		inf = 255 * pow((i + 0.5) / 255.5 * c, g) + 0.5;
		gammatable_256[i] = CLAMP(0, inf, 255);
	}
}

void Image_Apply_Gamma_To_RGBA_Pixels (byte *data, int width, int height, byte gammatable_256[])
{
	int imagesize = width * height;
	int i;
	for (i = 0 ; i < imagesize ; i++)
	{
		data[4*i+0] = gammatable_256[data[4*i+0]];
		data[4*i+1] = gammatable_256[data[4*i+1]];
		data[4*i+2] = gammatable_256[data[4*i+2]];
	}
}


///////////////////////////////////////////////////////////////////////////////
//  PALETTE REPLACEMENT
///////////////////////////////////////////////////////////////////////////////

static byte *fastpal;
static byte	 fastblack;

void Image_Quick_Palette_256_Free (void)
{
	free (fastpal);
	fastpal = NULL;
}

void Image_Quick_Palette_256_Alloc (int black_index)
{
	// Check to see if it was already allocated
	if (fastpal)
	{
		Core_Printf ("Quick palette wasn't closed. Freeing ...\n");
		Image_Quick_Palette_256_Free ();
	}

	// Baker: Allocate a buffer for all possible RGB combinations, calloc zeros the memory
	fastpal = (byte*)calloc (256*256*256, 1);
	fastblack = (byte)black_index;
}


byte Palette_Best_Color_Index_For_Pixel (unsigned mypixel, byte palette768[], int numpalcolors)
{
	static byte red, green, blue; //, alpha;
	static int red_distance, green_distance, blue_distance;
	static int color_distance, best_color_index;
	static unsigned best_color_distance;
	static byte *current_palette_index;
	static int i;

	unsigned rgbpix = mypixel & 0xFFFFFF; // Remove alpha from the equation

	byte fasteval = fastpal ? fastpal[rgbpix] : 0;

	if (fasteval)
	{
		if (mypixel == 0)
			return fastblack;

		return fasteval;
	}

	red		= (mypixel      )  & 0xFF;
	green	= (mypixel >> 8 )  & 0xFF;
	blue	= (mypixel >> 16)  & 0xFF;
//	alpha	= (mypixel >> 24)  & 0xFF;

	best_color_index		=  -1;
	best_color_distance		=  0 - 1; // Should be a big number
	current_palette_index	= palette768;

	for (i = 0; i < numpalcolors; i++, current_palette_index += 3 )
	{
		red_distance	= red   - current_palette_index[0];
		green_distance	= green - current_palette_index[1];
		blue_distance	= blue  - current_palette_index[2];
		color_distance	= SQUARED(red_distance) + SQUARED(green_distance) + SQUARED(blue_distance); // Could sqrt this but no reason to do so

		if ((unsigned int)color_distance < best_color_distance)
		{
			best_color_distance	= color_distance;
			best_color_index		= i;

			if (best_color_distance == 0)
				break; // Can't beat an exact match
		}
	}

	if (fastpal)
		fastpal[rgbpix] = (byte)best_color_index;

	return (byte)best_color_index;
}

void Image_Convert_Palette_To_Palette (byte *pixels_indexed, int len, byte old_palette_768[], byte new_palette_768[] )
{
	byte	newpal[256]; // For each HL color, we will fill this in
	int i, j;
	unsigned mycolor;
	byte	qpal;

	for (i = 0, j = 0; i < 256; i++, j+=3)
	{
		mycolor = ((unsigned)old_palette_768[j + 0] << 0)
				+ ((unsigned)old_palette_768[j + 1] << 8)
				+ ((unsigned)old_palette_768[j + 2] << 16);

		newpal[i] = qpal = Palette_Best_Color_Index_For_Pixel (mycolor, new_palette_768, 224);
	}

	// Baker: I think Half-Life color #255 is special like Quake
	newpal[255] = 255;

	// Now convert the pixels
	for (i = 0; i < len; i ++)
	{
		pixels_indexed[i] = newpal[pixels_indexed[i]];
	}
}

byte *Image_Buffer_RGBA_To_Palette_Alpha_Threshold_Alloc (byte *pixels_rgba, int width, int height, byte palette_768[], int alpha_threshold)
{
	int size = width * height;
	byte *buf = malloc (size);
	byte *dst;
	unsigned *src;
	int i;
	unsigned mypixel;
	byte red, green, blue, alpha;

	for (i = 0, src = (unsigned*) pixels_rgba, dst = buf; i < size; i++, dst++, src++)
	{
		mypixel = *src;
		red		= (mypixel      )  & 0xFF;
		green	= (mypixel >> 8 )  & 0xFF;
		blue	= (mypixel >> 16)  & 0xFF;
		alpha	= (mypixel >> 24)  & 0xFF;

		if (alpha <= alpha_threshold)
			*dst = 255; // Quake transparent color
		else
			*dst = Palette_Best_Color_Index_For_Pixel (*src, palette_768, 255); // First 255 colors, not 256
	}
	return buf;
}



unsigned int *fastpal_rgb4;
void Image_Quick_Palette_RGBA_Free (void)
{
	free (fastpal_rgb4);
	fastpal_rgb4 = NULL;
}

void Image_Quick_Palette_RGBA_Alloc (void)
{
	// Check to see if it was already allocated
	if (fastpal_rgb4)
	{
		Core_Printf ("Quick palette wasn't closed. Freeing ...\n");
		Image_Quick_Palette_RGBA_Free ();
	}

	// Baker: Allocate a buffer for all possible RGB combinations, calloc zeros the memory
	fastpal_rgb4 = (unsigned int*)calloc (256*256*256, 4);
}


///////////////////////////////////////////////////////////////////////////////
//  IMAGE: BUFFERED READ (STATIC)
///////////////////////////////////////////////////////////////////////////////


typedef struct stdio_buffer_s 
{
	FILE *f;
	unsigned char buffer[1024];
	int size;
	int pos;
} stdio_buffer_t;

static stdio_buffer_t *Buf_Alloc(FILE *f)
{
	stdio_buffer_t *buf = core_calloc (1, sizeof(stdio_buffer_t));
	buf->f = f;
	return buf;
}

static void Buf_Free(stdio_buffer_t *buf)
{
	core_free (buf);
}

static inline int Buf_GetC(stdio_buffer_t *buf)
{
	if (buf->pos >= buf->size)
	{
		buf->size = fread(buf->buffer, 1, sizeof(buf->buffer), buf->f);
		buf->pos = 0;

		if (buf->size == 0)
			return EOF;
	}

	return buf->buffer[buf->pos++];
}


///////////////////////////////////////////////////////////////////////////////
//  IMAGE: TGA with bgra and rgba support
///////////////////////////////////////////////////////////////////////////////

typedef struct targaheader_s 
{
	unsigned char 	id_length, colormap_type, image_type;
	unsigned short	colormap_index, colormap_length;
	unsigned char	colormap_size;
	unsigned short	x_origin, y_origin, width, height;
	unsigned char	pixel_size, attributes;
} targaheader_t;

#define TARGAHEADERSIZE 18 //size on disk

targaheader_t targa_header;

static int fgetLittleShort (FILE *f)
{
	byte	b1, b2;

	b1 = fgetc(f);
	b2 = fgetc(f);

	return (short)(b1 + b2*256);
}

/* unused
static int fgetLittleLong (FILE *f)
{
	byte	b1, b2, b3, b4;

	b1 = fgetc(f);
	b2 = fgetc(f);
	b3 = fgetc(f);
	b4 = fgetc(f);

	return b1 + (b2<<8) + (b3<<16) + (b4<<24);
}
*/


cbool Image_Save_TGA (const char *path_to_file, byte *data, int width, int height, int bits_per_pixel, cbool upsidedown)
{
	int		i, size, temp, bytes;
	byte	header[TARGAHEADERSIZE];
	FILE *f;
	
	f = core_fopen_write (path_to_file, "wb");
	
	if (!f)
		return false;

	memset (&header, 0, TARGAHEADERSIZE);
	header[2] = 2; // uncompressed type
	header[12] = width&255;
	header[13] = width>>8;
	header[14] = height&255;
	header[15] = height>>8;
	header[16] = bits_per_pixel; // pixel size
	if (upsidedown)
		header[17] = 0x20; //upside-down attribute

	// swap red and blue bytes
	bytes = bits_per_pixel / 8;
	size = width * height * bytes;

	for (i = 0; i < size; i += bytes)
	{
		temp = data[i];
		data[i] = data[i+2];
		data[i+2] = temp;
	}

	fwrite (&header, 1, TARGAHEADERSIZE, f);
	fwrite (data, 1, size, f);

	core_fclose (f);

	return true;
}


byte *Image_Load_TGA_FileHandle (FILE *fin, int *width, int *height, malloc_fn_t Malloc_Fn, const char* description)
{
	int				columns, rows, numPixels;
	byte			*pixbuf;
	int				row, column;
	byte			*targa_rgba;
	int				realrow; //johnfitz -- fix for upside-down targas
	cbool			upside_down; //johnfitz -- fix for upside-down targas
	stdio_buffer_t	*buf;

	targa_header.id_length = fgetc(fin);
	targa_header.colormap_type = fgetc(fin);
	targa_header.image_type = fgetc(fin);

	targa_header.colormap_index = fgetLittleShort(fin);
	targa_header.colormap_length = fgetLittleShort(fin);
	targa_header.colormap_size = fgetc(fin);
	targa_header.x_origin = fgetLittleShort(fin);
	targa_header.y_origin = fgetLittleShort(fin);
	targa_header.width = fgetLittleShort(fin);
	targa_header.height = fgetLittleShort(fin);
	targa_header.pixel_size = fgetc(fin);
	targa_header.attributes = fgetc(fin);

	if (targa_header.image_type!=2 && targa_header.image_type!=10)
		Core_Error ("Image_LoadTGA: %s is not a type 2 or type 10 targa", description);

	if (targa_header.colormap_type !=0 || (targa_header.pixel_size!=32 && targa_header.pixel_size!=24))
		Core_Error ("Image_LoadTGA: %s is not a 24bit or 32bit targa", description);

	columns = targa_header.width;
	rows = targa_header.height;
	numPixels = columns * rows;
	upside_down = !(targa_header.attributes & 0x20); //johnfitz -- fix for upside-down targas

	targa_rgba = (byte *)Malloc_Fn (numPixels*4);

	if (targa_header.id_length != 0)
		fseek (fin, targa_header.id_length, SEEK_CUR);  // skip TARGA image comment

	buf = Buf_Alloc(fin);

	if (targa_header.image_type == 2) // Uncompressed, RGB images
	{
		for(row=rows-1; row>=0; row--)
		{
			//johnfitz -- fix for upside-down targas
			realrow = upside_down ? row : rows - 1 - row;
			pixbuf = targa_rgba + realrow*columns*4;
			//johnfitz
			for(column=0; column<columns; column++)
			{
				unsigned char red,green,blue,alphabyte;
				switch (targa_header.pixel_size)
				{
				case 24:
					blue = Buf_GetC(buf);
					green = Buf_GetC(buf);
					red = Buf_GetC(buf);
					*pixbuf++ = red;
					*pixbuf++ = green;
					*pixbuf++ = blue;
					*pixbuf++ = 255;
					break;
				case 32:
					blue = Buf_GetC(buf);
					green = Buf_GetC(buf);
					red = Buf_GetC(buf);
					alphabyte = Buf_GetC(buf);
					*pixbuf++ = red;
					*pixbuf++ = green;
					*pixbuf++ = blue;
					*pixbuf++ = alphabyte;
					break;
				}
			}
		}
	}
	else if (targa_header.image_type==10) // Runlength encoded RGB images
	{
		unsigned char red,green,blue,alphabyte,packetHeader,packetSize,j;
		for(row = rows - 1; row >= 0; row--)
		{
			//johnfitz -- fix for upside-down targas
			realrow = upside_down ? row : rows - 1 - row;
			pixbuf = targa_rgba + realrow * columns * 4;
			//johnfitz
			for(column = 0; column < columns; )
			{
				packetHeader=Buf_GetC(buf);
				packetSize = 1 + (packetHeader & 0x7f);
				if (packetHeader & 0x80) // run-length packet
				{
					switch (targa_header.pixel_size)
					{
					case 24:
						blue = Buf_GetC(buf);
						green = Buf_GetC(buf);
						red = Buf_GetC(buf);
						alphabyte = 255;
						break;
					case 32:
						blue = Buf_GetC(buf);
						green = Buf_GetC(buf);
						red = Buf_GetC(buf);
						alphabyte = Buf_GetC(buf);
						break;
					default: /* avoid compiler warnings */
						blue = red = green = alphabyte = 0;
					}

					for(j = 0;j < packetSize; j++)
					{
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = alphabyte;
						column++;
						if (column == columns) // run spans across rows
						{
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut;
							//johnfitz -- fix for upside-down targas
							realrow = upside_down ? row : rows - 1 - row;
							pixbuf = targa_rgba + realrow*columns*4;
							//johnfitz
						}
					}
				}
				else // non run-length packet
				{
					for(j=0;j<packetSize;j++)
					{
						switch (targa_header.pixel_size)
						{
						case 24:
							blue = Buf_GetC(buf);
							green = Buf_GetC(buf);
							red = Buf_GetC(buf);
							*pixbuf++ = red;
							*pixbuf++ = green;
							*pixbuf++ = blue;
							*pixbuf++ = 255;
							break;
						case 32:
							blue = Buf_GetC(buf);
							green = Buf_GetC(buf);
							red = Buf_GetC(buf);
							alphabyte = Buf_GetC(buf);
							*pixbuf++ = red;
							*pixbuf++ = green;
							*pixbuf++ = blue;
							*pixbuf++ = alphabyte;
							break;
						default: /* avoid compiler warnings */
							blue = red = green = alphabyte = 0;
						}
						column++;
						if (column == columns) // pixel packet run spans across rows
						{
							column = 0;
							if (row > 0)
								row--;
							else
								goto breakOut;
							//johnfitz -- fix for upside-down targas
							realrow = upside_down ? row : rows - 1 - row;
							pixbuf = targa_rgba + realrow*columns*4;
							//johnfitz
						}
					}
				}
			}
			breakOut:;
		}
	}

	Buf_Free(buf);
	core_fclose(fin);

	*width = (int)(targa_header.width);
	*height = (int)(targa_header.height);
	return targa_rgba;
}


///////////////////////////////////////////////////////////////////////////////
//  IMAGE: PCX (READ ONLY)
///////////////////////////////////////////////////////////////////////////////

typedef struct
{
    char			signature;
    char			version;
    char			encoding;
    char			bits_per_pixel;
    unsigned short	xmin,ymin,xmax,ymax;
    unsigned short	hdpi,vdpi;
    byte			colortable[48];
    char			reserved;
    char			color_planes;
    unsigned short	bytes_per_line;
    unsigned short	palette_type;
    char			filler[58];
} pcxheader_t;


byte * Image_Load_PCX_FileHandle (FILE *fin, int *width, int *height, size_t filelen, malloc_fn_t Malloc_Fn, const char* description)
{
	pcxheader_t	pcx;
	int			x, y, w, h, readbyte, runlength, start;
	byte		*p, *data;
	byte		palette[768];
	stdio_buffer_t  *buf;

	start = ftell (fin); //save start of file (since we might be inside a pak file, SEEK_SET might not be the start of the pcx)

	fread (&pcx, sizeof(pcx), 1, fin);
	pcx.xmin = (unsigned short)pcx.xmin;
	pcx.ymin = (unsigned short)pcx.ymin;
	pcx.xmax = (unsigned short)pcx.xmax;
	pcx.ymax = (unsigned short)pcx.ymax;
	pcx.bytes_per_line = (unsigned short)pcx.bytes_per_line;

	if (pcx.signature != 0x0A)
		Core_Error ("'%s' is not a valid PCX file", description);

	if (pcx.version != 5)
		Core_Error ("'%s' is version %i, should be 5", description, pcx.version);

	if (pcx.encoding != 1 || pcx.bits_per_pixel != 8 || pcx.color_planes != 1)
		Core_Error ("'%s' has wrong encoding or bit depth", description);

	w = pcx.xmax - pcx.xmin + 1;
	h = pcx.ymax - pcx.ymin + 1;

	data = (byte *)Malloc_Fn((w * h + 1) * 4); //+1 to allow reading padding byte on last line

	//load palette
	fseek (fin, start + filelen - 768, SEEK_SET);
	fread (palette, 1, 768, fin);

	//back to start of image data
	fseek (fin, start + sizeof(pcx), SEEK_SET);

	buf = Buf_Alloc(fin);

	for (y=0; y<h; y++)
	{
		p = data + y * w * 4;

		for (x=0; x < pcx.bytes_per_line; ) //read the extra padding byte if necessary
		{
			readbyte = Buf_GetC(buf);

			if(readbyte >= 0xC0)
			{
				runlength = readbyte & 0x3F;
				readbyte = Buf_GetC(buf);
			}
			else
				runlength = 1;

			while(runlength--)
			{
				p[0] = palette[readbyte*3];
				p[1] = palette[readbyte*3+1];
				p[2] = palette[readbyte*3+2];
				p[3] = 255;
				p += 4;
				x++;
			}
		}
	}

	Buf_Free(buf);
	core_fclose(fin);

	*width = w;
	*height = h;
	return data;
}

///////////////////////////////////////////////////////////////////////////////
//  IMAGE: PNG
///////////////////////////////////////////////////////////////////////////////


#include "lodepng.h"

cbool Image_Save_PNG (const char *path_to_file, byte *data, int width, int height)
{
	// Baker: We are assuming we have write access to the buffer
	{
		unsigned error = lodepng_encode32_file(path_to_file, data, width, height);
		if (error)
		{
			Core_Printf ("Error writing png\n");
			return 0;
		}
	}
	return 1;
}




