/*
Copyright (C) 2012-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
// memchain.c -- memory chain for temp allocs

#include "core.h"
#include "memchain.h"

///////////////////////////////////////////////////////////////////////////////
//  LOCAL SETUP
///////////////////////////////////////////////////////////////////////////////

#define MLOCAL(_m) struct mlocal *_m = (struct mlocal *) me->_local;

// Begin ...

static const char * const _tag = "memchain";  	// TAG
#define mobj_t memchain_t						// object type
#define mlocal _memchain_local					// private data struct

typedef struct mitem_s
{
	struct mitem_s *next, *prev;
	size_t bytes;
	void *data;
} mitem_t;

typedef struct memtrack_s
{
	mitem_t		*head;
	mitem_t		*tail;
	int			count;
	int			bytes;
} memtrack_t;

struct mlocal							// private data
{
	// Make this
	memtrack_t tracker;
};


///////////////////////////////////////////////////////////////////////////////
//  INTERNAL FUNCTIONS
///////////////////////////////////////////////////////////////////////////////

static void sLink (memtrack_t *track, mitem_t *item)
{
	// Assume head if not taken
	if (track->head) track->head = item;
	
	// Add to tail
	if (track->tail) 
	{
		track->tail->next = item;
		item->prev = track->tail;
	}

	// me->next = NULL;   // implicit
	track->count ++;
	track->bytes += item->bytes;
}


static void sUnlink (memtrack_t *track, mitem_t *item)
{
	if (item->prev) 
		item->prev->next = item->next;
	else track->head = NULL; // I was head

	if (item->next) 
		item->next->prev = item->prev;
	else track->tail = NULL; // I was tail

	track->count --;
	track->bytes -= item->bytes;
}


static mitem_t *sFindLink (memtrack_t *track, void *ptr)
{
	mitem_t *cur;

	for (cur = track->head; cur; cur = cur->next)
		if (cur->data == ptr)
			return cur;

	return NULL;
}


///////////////////////////////////////////////////////////////////////////////
//  PUBLIC OBJECT FUNCTIONS:
///////////////////////////////////////////////////////////////////////////////

static void *Malloc (mobj_t *me, size_t len)
{
	MLOCAL(m)

	mitem_t *item = core_malloc (sizeof(mitem_t));
	item->bytes = len;
	item->data = core_malloc (len);

	sLink (&m->tracker, item);
	return item->data;
}

static void *Memdup (mobj_t *me, const void *src, size_t len)
{
	MLOCAL(m)

	mitem_t *item = core_malloc (sizeof(mitem_t));
	item->bytes = len;
	item->data = core_malloc (len);

	memcpy (item->data, src, len);
	sLink (&m->tracker, item);
	return item->data;
}


static void *Calloc (mobj_t *me, size_t len, size_t n)
{
	MLOCAL(m)

	mitem_t *item = core_malloc (sizeof(mitem_t));
	item->bytes = len * n;
	item->data = core_calloc (len, n);

	sLink (&m->tracker, item);
	return item->data;
}

static char *Strdup (mobj_t *me, const char *s)
{
	MLOCAL(m)

	mitem_t *item = core_malloc (sizeof(mitem_t));
	item->bytes = strlen (s) + 1;
	item->data = core_strdup (s);
	
	sLink (&m->tracker, item);
	return (char *)item->data;
}

static void *Free (mobj_t *me, void *ptr)
{
	MLOCAL(m)

	mitem_t *item = sFindLink (&m->tracker, ptr);
	
	if (item == NULL)
		Core_Error ("Missing Link");

	sUnlink (&m->tracker, item);

	core_free (item->data);
	core_free (item);
	return NULL;
}

static void *Realloc (mobj_t *me, void *ptr, size_t len)
{
	MLOCAL(m)

	mitem_t *item = sFindLink (&m->tracker, ptr);

	if (item == NULL)
		Core_Error ("Missing Link");
	
	sUnlink (&m->tracker, item);

	item->bytes = len;
	item->data = realloc (item->data, len);

	sLink (&m->tracker, item);

	return item->data;
}

static void Flush (mobj_t *me)
{
	MLOCAL(m)

	memtrack_t *track = &m->tracker;
	mitem_t *cur;

	for (cur = track->head; cur; cur = cur->next)
	{
		sUnlink (&m->tracker, cur);
		core_free (cur->data);
		core_free (cur);
	}
}

static cbool Initialize (mobj_t *me)
{
	MLOCAL(m)

	return true;
}


///////////////////////////////////////////////////////////////////////////////
//  PUBLIC GLOBAL FUNCTIONS:  Creation and freeing of object
///////////////////////////////////////////////////////////////////////////////


// Shutdown and Free
static void *Shutdown (mobj_t *me)
{
	MLOCAL(m)

	Flush (me);

	core_free (me->_local);
	core_free (me);

	return NULL;
}



memchain_t *Memchain_Instance (void)
{
	cbool result;

	mobj_t *nobj = core_calloc (sizeof(mobj_t), 1);
	nobj->_local = core_calloc (sizeof(struct mlocal), 1);

	// Hook up functions
	#define FUNCTION_HOOKUP(_x) nobj->_x = _x

	FUNCTION_HOOKUP(Malloc);
	FUNCTION_HOOKUP(Calloc);
	FUNCTION_HOOKUP(Strdup);
	FUNCTION_HOOKUP(Free);

	FUNCTION_HOOKUP(Realloc);
	FUNCTION_HOOKUP(Memdup);

	FUNCTION_HOOKUP(Flush);
	#undef FUNCTION_HOOKUP

	// Run initializer
	result = Initialize (nobj);

	if (!result)
	{
		nobj = Shutdown (nobj);
	}

	// Required information
	OBJ_REQUIRED_HOOKUP(nobj) // Sets parent, _tag, Shutdown

	return nobj;
}

#undef mobj_t 
#undef mlocal

