/*
Copyright (C) 2012-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
// sys_register.c -- tracks startup and shutdown

#include "core.h"
#include "corex.h"
#include "sys_register.h" // Courtesy include

///////////////////////////////////////////////////////////////////////////////
//  LOCAL SETUP
///////////////////////////////////////////////////////////////////////////////

#define MLOCAL(_m) struct mlocal *_m = (struct mlocal *) me->_local;

// Begin ...

static const char * const _tag = "sysreg";  	// TAG
#define mobj_t sysreg_t						// object type
#define mlocal _sysreg_local					// private data struct
#define mitem_t _sysreg_item_t
#define mitem_s _sysreg_item_s
#define mtracker_s _sysreg_track_t
#define mtracker_t _sysreg_track_s

typedef struct mitem_s
{
	struct mitem_s *next, *prev;
	sysobj_t	*object;
	const char	*objname;
} mitem_t;

typedef struct mtracker_s
{
	mitem_t		*head;
	mitem_t		*tail;
	int			count;
} mtracker_t;

struct mlocal							// private data
{
	mtracker_t	tracker;	// Tracks subsystems
	memchain_t *memchain;	// Stores our memory allocations for easy disposal
};


///////////////////////////////////////////////////////////////////////////////
//  INTERNAL FUNCTIONS
///////////////////////////////////////////////////////////////////////////////

static void sLink (mtracker_t *track, mitem_t *item)
{
	// Assume head if not taken
	if (!track->head) track->head = item;
	
	// Add to tail
	if (track->tail) 
	{
		track->tail->next = item;
		item->prev = track->tail;
	}

	// me->next = NULL;   // implicit
	track->count ++;
}


static void sUnlink (mtracker_t *track, mitem_t *item)
{
	if (item->prev) 
		item->prev->next = item->next;
	else track->head = NULL; // I was head

	if (item->next) 
		item->next->prev = item->prev;
	else track->tail = NULL; // I was tail

	track->count --;
}


static mitem_t *sFindLink (mtracker_t *track, sysobj_t *object)
{
	mitem_t *cur;

	for (cur = track->head; cur; cur = cur->next)
		if (cur->object == object)
			return cur;

	return NULL;
}


///////////////////////////////////////////////////////////////////////////////
//  PUBLIC INFORMATIONAL OBJECTS
///////////////////////////////////////////////////////////////////////////////


static void Report (mobj_t *me)
{
	MLOCAL(m)

	mitem_t *cur;
	int count;

	Core_Printf ("System Register\n-----\n");
	// Print A to Z
	for (cur = m->tracker.head, count = 0; cur; cur = cur->next, count ++)
		Core_Printf ("%03i: (object: %p) Type: \"%s\"\n", count, cur->object, cur->objname);

	Core_Printf ("--Found %i -----\n", count);
}

static int Report_Cmd (sysobj_t *me, lparse_t *line)
{
	Report ((mobj_t *) me);
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//  PUBLIC OBJECT FUNCTIONS:
///////////////////////////////////////////////////////////////////////////////

static void *sObjectShutdownItem (mobj_t *me, mitem_t *cur)
{
	MLOCAL (m)

	// Verify
	if (strcmp (cur->objname, cur->object->_cname))
		Core_Error ("Shutdown object name does not match");

	// Run shutdown function
	cur->object->Shutdown (cur->object);

	// Unlink from our list
	sUnlink (&m->tracker, cur);

	// Deallocate our copy of the name
	m->memchain->Free (m->memchain, (void *)cur->objname);

	// Deallocate link item it using memory chain
	m->memchain->Free (m->memchain, cur);

	return NULL;
}

// Close one subsystem
static void *ObjectShutdown (mobj_t *me, sysobj_t *object)
{	
	// If object is NULL that would be disturbing
	if (object)
	{
		MLOCAL(m)
		
		// Locate it in the list
		mitem_t *cur = sFindLink (&m->tracker, object);

		if (!cur)
			Core_Error ("Unable to shutdown object because it is not registered!");
		
		return sObjectShutdownItem (me, cur);
	}
	return NULL;
}



static void *ObjectRegister (mobj_t *me, sysobj_t *object)
{
	// Only if registration succeeded (failed is NULL)
	if (object)
	{
		MLOCAL(m)

		// Use the memory chain to allocate a subsystems item item
		mitem_t *item = m->memchain->Calloc (m->memchain, sizeof(mitem_t), 1);

		// Fill in the object
		item->object = object;
		item->objname = m->memchain->Strdup (m->memchain, object->_cname);

		// Link it in
		sLink (&m->tracker, item);

		if (!strcmp(item->objname, "cmdsys"))
		{
			// If it is a command system, register our commands with it
			static cmdinfo_t commands[] =
			{
				"subsystems", Report_Cmd, "List of registered systems"
			}; ARRAY_SIZE_STATIC(commands);
			int i;
			cmdsys_t *cmdobj = (cmdsys_t *)object;

			// Add our commands
			for (i = 0; i < _numcommands; i ++)
				cmdobj->CommandAdd (cmdobj, (sysobj_t *)me, &commands[i]);
		}
	}
	return object;
}



static cbool Initialize (mobj_t *me)
{
	MLOCAL(m)

	// Anything we do here
	m->memchain = Memchain_Instance ();

	return true;
}


///////////////////////////////////////////////////////////////////////////////
//  PUBLIC GLOBAL FUNCTIONS:  Creation and freeing of object
///////////////////////////////////////////////////////////////////////////////


// Shutdown and Free
static void *Shutdown (mobj_t *me)
{
	MLOCAL(m)

	mitem_t *cur;

	// Shutdown in reverse order of startup
	for (cur = m->tracker.tail; cur; cur = cur->prev)
		sObjectShutdownItem (me, cur);
		
	m->memchain->Flush (m->memchain);

	core_free (me->_local);
	core_free (me);

	return NULL;
}


sysreg_t *SysReg_Instance (void)
{
	cbool result;

	mobj_t *nobj = core_calloc (sizeof(mobj_t), 1);
	nobj->_local = core_calloc (sizeof(struct mlocal), 1);

	// Hook up functions
	#define FUNCTION_HOOKUP(_x) nobj->_x = _x

	FUNCTION_HOOKUP(ObjectRegister);
	FUNCTION_HOOKUP(ObjectShutdown);
	FUNCTION_HOOKUP(Report);
//	FUNCTION_HOOKUP(StringCommand);

	#undef FUNCTION_HOOKUP

	// Run initializer
	result = Initialize (nobj);

	if (!result)
	{
		nobj = Shutdown (nobj);
	}

	// Required information
	OBJ_REQUIRED_HOOKUP(nobj) // Sets parent, _tag, Shutdown

	return nobj;
}

#undef mobj_t 
#undef mlocal

	